///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
/// =========================================================================
#include "rheolef/d_dn.h"

namespace rheolef {
using namespace std;
using namespace ublas;

/* 
   Let
     S in Omega  : a side of the surface Omega
     K in Lambda : a element of the background domain Lambda
     we assume that S is a side of K

   d_dn uses the two piola transformations:
        F : hat_S   ----> S
            hat_x   |---> x = F(hat_S)

	G : tilde_K ----> K
            tilde_x |---> x = G(tilde_x)

   The gradient is transformed as:
        grad = inv(DG)^T*tilde_grad
   A side tilde_S of tilde_K is transformed by G as the side S of K.
   The corresponding measures are d_tilde_s and ds with:
	ds = Js*d_tilde_s
   The unit outward normal tilde_n on hat_S is related to n on S as (Rav-1981, p.24):
        n  = (det(DG)/Js)*DG^{-T}*tilde_n
   Let us introduce the matrix:
        CG = det(DG)*inv(DG)*inv(DG)^T
   Then
        n.grad = (1/Js)*tilde_n.(CG*tilde_grad)

   This expression is summed over S by using a quadrature formula.
   The quadrature formula is defined on the reference element hat_S 

        int_S (n.grad)v lambda ds
        = sum_q {(1/Js)*tilde_n.(CG*tilde_grad(tilde_v))}(tilde_xq)
                hat_lambda(hat_xq)
                |DF(hat_xq)| hat_wq

   While the second part is defined for hat_x,
   the first part of the expression, related to v, is defined for
        tilde_x = G^{-1}(x) = G^{-1}(F(hat_x))
   Thus, we have to evaluate DG at all tilde_xq, the transformed points.

   Implementation notes:
   1) |DF(hat_xq)|/Js(tilde_xq) refers to the relation:
	ds = Js*d_tilde_s = |DF(hat_xq)|*d_hat_s
   It expresses the determinant |DH| of the jacobian of the transformation
     H: hat_S	----> tilde_S
	hat_x   |---> tilde_x = H(hat_x)
   and we have H = F o G^{-1}. Then
	det(DH) = |DF(hat_xq)|/Js = meas(tilde_S)/meas(hat_S)
   It depends only of the local side index
   of tilde_S on the reference element tilde_K.
   For instance, for a triangle K, and and edge S:
         loc_isid	|DH| = |DF(hat_xq)|/Js
         0		1
         1		sqrt(2)
         2		1

   2) The normal tilde_n is constant and can be also computed 
   outside of the loop: 
   It depends only of the local side index (loc_isid)
   of tilde_S on the reference element tilde_K.
   For instance, for a triangle K, and and edge S:
         loc_isid	tilde_n
         0		[ 0,-1]
         1		[ 1, 1]/sqrt(2)
         2		[-1, 0]

   3) All points tilde_xq are defined one time for all on
   the reference element tilde_K as the set of quadrature points for all faces.
   Thus, polynoms for DG (piola basis) and the tilde_v basis
   can be evaluated one time for all on these fixed points.

   This last point is not yet implemented, as it requires to define the poinset of
   merged quadrature nodes of all faces of an element, and then to address
   the good 'q' index inside this pointset.

 */

template<class T, class M>
void 
d_dn<T,M>::operator() (const geo_element& S, ublas::matrix<T>& m) const
{
  check_macro (! base::is_on_band(), "unsupported banded level set extension");

  const geo_basic<T,M>& Lambda = base::_omega.get_background_geo();
  size_type S_map_d = S.dimension();
  size_type K_map_d = S_map_d+1;
  size_type        d = base::coordinate_dimension();
  // get the element K containing the side S:
  size_type dis_ie = S.master(0);
  check_macro (dis_ie != std::numeric_limits<size_type>::max(), 
	"unexpected side S="<<S<<" without master element");
  const geo_element& K = base::_omega.dis_get_geo_element (K_map_d, dis_ie);
  reference_element hat_S = S;
  reference_element tilde_K = K;
  size_type nx = base:: get_first_basis().size(tilde_K);
  size_type ny = base::get_second_basis().size(hat_S);
  m.resize (ny, nx);
  m.clear();
  std::vector<size_type> dis_inod_S;
  std::vector<size_type> dis_inod_K;
        Lambda.dis_inod (K, dis_inod_K);
  base::_omega.dis_inod (S, dis_inod_S);
  size_type loc_isid, shift;
  K.get_side_informations (S, loc_isid, shift);
  Float det_DH  = tilde_K.side_measure(loc_isid)/measure(hat_S);
  point tilde_n;  tilde_K.side_normal (loc_isid, tilde_n);
  typename quadrature<T>::const_iterator
    first = base::_quad.begin(hat_S),
    last  = base::_quad.end  (hat_S);
  tensor_basic<T> DF, DG, invDG, CG_wq;
  for (size_type q = 0; first != last; first++, q++) {
    jacobian_piola_transformation (base::_omega, base::_piola_table, S, dis_inod_S, q, DF);
    point_basic<T> xq = piola_transformation (base::_omega, base::_piola_table, hat_S, dis_inod_S, q);
    point_basic<T> tilde_xq = inverse_piola_transformation (Lambda, tilde_K, dis_inod_K, xq);
    jacobian_piola_transformation (Lambda, base::_piola_table, tilde_K, dis_inod_K, tilde_xq, DG);
    T det_DG = det_jacobian_piola_transformation (DG, d, K_map_d);
    invDG = pseudo_inverse_jacobian_piola_transformation (DG, d, K_map_d);
    T wq = base::is_weighted() ? base::weight(S,q) : T(1);
    wq *= base::weight_coordinate_system (S, dis_inod_S, q);
    wq *= det_DH;
    wq *= (*first).w;
    CG_wq = wq*det_DG*invDG*trans(invDG); // TODO: use reduced sizes K_map_d
    base::_bx_table.evaluate_grad (tilde_K, tilde_xq); // TODO: define a custom side pointset (see note #3)
    base::_by_table.evaluate      (hat_S,   q);
    typename base::quad_const_iterator_grad
      first_grad_v = base::_bx_table.begin_grad (),
       last_grad_v = base::_bx_table.end_grad   ();
    typename base::quad_const_iterator
      first_lambda = base::_by_table.begin(),
       last_lambda = base::_by_table.end  ();
    for (size_type j = 0; first_grad_v != last_grad_v; first_grad_v++, j++) {
      const point_basic<T>& grad_v_j = *first_grad_v;
      T value_j = dot (tilde_n, (CG_wq*grad_v_j)); // TODO: use reduced sizes K_map_d
      typename base::quad_const_iterator lambda = first_lambda;
      for (size_type i = 0; lambda != last_lambda; lambda++, i++) {
        T lambda_i = *lambda;
	m(i,j) += value_j*lambda_i;
      }
    }
  }
}
template<class T, class M>
void
d_dn<T,M>::initialize () const
{
  base::set_n_derivative(1);
  check_macro (
        space_constant::scalar == base::get_second_space().valued_tag(),
        "unsupported non-scalar second space for `d_dn' form");
  check_macro (
        space_constant::scalar == base::get_first_space().valued_tag(),
        "unsupported non-scalar first space for `d_dn' form");
  base::_omega.neighbour_guard(); // uses S.master(i), a neighbour connnectivity
}
template<class T, class M>
void
d_dn<T,M>::first_dis_idof (const geo_element& S, std::vector<size_type>& dis_jdx) const
{
  size_type dis_ie = S.master(0);
  check_macro (dis_ie != std::numeric_limits<size_type>::max(), 
	"unexpected side S="<<S<<" without master element");
  size_type K_map_d = S.dimension() + 1;
  const geo_element& K = base::_omega.dis_get_geo_element (K_map_d, dis_ie);
  base::get_first_space().dis_idof (K, dis_jdx);
#ifdef TO_CLEAN
  for (size_t i = 0, n = dis_jdx.size(); i < n; i++) {
    std::cout << "dis_jdx["<<i<<"]="<< dis_jdx[i] << std::endl;
  }
#endif // TO_CLEAN
}
template<class T, class M>
void
d_dn<T,M>::second_dis_idof (const geo_element& S, std::vector<size_type>& dis_idy) const
{
  base::get_second_space().dis_idof (S, dis_idy);
#ifdef TO_CLEAN
  for (size_t i = 0, n = dis_idy.size(); i < n; i++) {
    std::cout << "dis_idy["<<i<<"]="<< dis_idy[i] << std::endl;
  }
#endif // TO_CLEAN
}
// ----------------------------------------------------------------------------
// instanciation in library
// ----------------------------------------------------------------------------
template class d_dn<Float,sequential>;

#ifdef _RHEOLEF_HAVE_MPI
template class d_dn<Float,distributed>;
#endif // _RHEOLEF_HAVE_MPI

} // namespace rheolef
