#ifndef _RHEOLEF_D_DN_H
#define _RHEOLEF_D_DN_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================

#include "rheolef/form_element.h"

namespace rheolef { 

/*Class:d_dn
NAME: @code{d_dn} -- normal derivative
@bfindex d_dn
@cindex  derivative
@clindex form
@clindex space
SYNOPSIS:
  @example
    	form (const space V, const space& M, "d_dn");
  @end example
DESCRIPTION:       
  @noindent
  Assembly the form (@pxref{form class}) associated to the normal
  derivative from the @code{V} finite element space
  to the @code{M} one.
  When @code{V} is a space defined on a domain Omega,
  assume that the @code{M} space is defined on the boundary
  Gamma of Omegai, where the unit outward normal is well-defined.
@iftex
@tex
  $$
      b(u,q) = \int_\Gamma
         { {\partial u} \over {\partial n}} \, q \, ds
  $$
@end tex
@end iftex
@ifnottex
  @example
                 /
                 |  d u
      b_i(u,q) = |  ----  q  ds
                 |  d n
                 / Gamma
  @end example
@end ifnottex

EXAMPLE:
  @noindent
  The following piece of code build the form:
  @example
   	geo omega ("square");
   	space V (omega, "P1");
   	space M (omega["boundary"], "P0");
   	form  b (V, M, "d_dn");
  @end example
AUTHOR: 
    LMC-IMAG, 38041 Grenoble cedex 9, France
    | Pierre.Saramito@imag.fr
SEE ALSO:
    "form"(3)
DATE:
    29 june 2012
End:
*/
template<class T, class M>
class d_dn: public form_element_rep<T,M> {
public:
    typedef form_element_rep<T,M>       base;
    typedef typename base::size_type    size_type;
    typedef typename base::geo_type     geo_type;
    typedef typename base::space_type   space_type;

    d_dn (const space_type& X, const space_type& Y, const geo_type& omega, 
        const quadrature_option_type& qopt)
	: form_element_rep<T,M> (X,Y,omega,qopt) { base::initialize_all(); }

    void operator() (const geo_element& S, ublas::matrix<T>& m) const;
    void  first_dis_idof (const geo_element& S, std::vector<size_type>& dis_jdx) const;
    void second_dis_idof (const geo_element& S, std::vector<size_type>& dis_idy) const;
    void initialize () const;
    bool is_symmetric () const { return false; }
};

}// namespace rheolef
#endif // _RHEOLEF_D_DN_H
