#ifndef _RHEOLEF_D_DX0_H
#define _RHEOLEF_D_DX0_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================

#include "rheolef/form_element.h"

namespace rheolef { 

/*Class:d_dx0
NAME: @code{d_dx}@var{i} -- derivatives
@bfindex d_dx0
@bfindex d_dx1
@bfindex d_dx2
@cindex  derivative
@clindex form
@clindex space
SYNOPSIS:
  @example
    	form (const space V, const space& M, "d_dx0");
    	form (const space V, const space& M, "d_dx1");
    	form (const space V, const space& M, "d_dx2");
  @end example
DESCRIPTION:       
  @noindent
  Assembly the form (@pxref{form class}) associated to a
  derivative operator from the @code{V} finite element space
  to the @code{M} one:
@iftex
@tex
  $$
      b_i(u,q) = \int_\Omega
         { {\partial u} \over {\partial x_i}} \, q \, dx,
	\ \ i\in \{ 0,1,2 \}
  $$
@end tex
@end iftex
@ifnottex
  @example
                 /
                 |  d u
      b_i(u,q) = |  ----  q  dx,  i = 0,1,2
                 |  d xi
                 / Omega
  @end example
@end ifnottex
  In the axisymetric @code{rz} case, the form @code{d_dx0} is defined by
@iftex
  $$
      b_0(u,q) = \int_\Omega 
         { {\partial u} \over {\partial r}} \, q \, r dr dz,
  $$
@end iftex
@ifnottex
  @example
                 /
                 |  d u
      b_0(u,q) = |  ---  q  r dr dz
                 |  d r
                 / Omega
  @end example
@end ifnottex

EXAMPLE:
  @noindent
  The following piece of code build the Laplacian form:
  @example
   	geo omega ("square");
   	space V (omega, "P1");
   	space M (omega, "P0");
   	form  b (V, M, "d_dx0");
  @end example
AUTHOR: 
    LMC-IMAG, 38041 Grenoble cedex 9, France
    | Pierre.Saramito@imag.fr
SEE ALSO:
    "form"(3)
DATE:
    28 november 1997
End:
*/
template<class T, class M>
class d_dx0: public form_element_rep<T,M> {
public:
    typedef form_element_rep<T,M>       base;
    typedef typename base::size_type    size_type;
    typedef typename base::geo_type     geo_type;
    typedef typename base::space_type   space_type;

    d_dx0 (const space_type& X, const space_type& Y, const geo_type& omega, 
        const quadrature_option_type& qopt)
	: form_element_rep<T,M> (X,Y,omega,qopt) { base::initialize_all(); }

    void operator() (const geo_element& K, ublas::matrix<T>& m) const;
    void initialize () const;
    bool is_symmetric () const { return false; }
};

}// namespace rheolef
#endif // _RHEOLEF_D_DX0_H
