#ifndef _RHEOLEF_FIELD_EVALUATE_H
#define _RHEOLEF_FIELD_EVALUATE_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================
// 
// evaluate a field on a predefined point set: hat_x[q], q=0..nq
// See also piola_transformation.h
//
#include "rheolef/field.h"
#include "rheolef/basis_on_pointset.h"
namespace rheolef { 

// -----------------------------------------------------
// scalar-valued case:
// -----------------------------------------------------
template<class T, class M, class BasisOnPointset>
T
field_evaluate (
  const field_basic<T,M>&       uh,
  const BasisOnPointset&        basis_on_pointset,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof,
  size_t                        q)
{
  typedef typename field_basic<T,M>::size_type size_type;
  T value = 0;
  size_type loc_idof = 0;
  for (typename BasisOnPointset::const_iterator
	  iter = basis_on_pointset.begin(hat_K, q),
          last = basis_on_pointset.end  (hat_K, q);
          iter != last;
          iter++, loc_idof++) {
    const T& cdof = *iter;
    const T& udof = uh.dis_dof (dis_idof[loc_idof]);
    value += cdof*udof;
  }
  return value;
}
// -----------------------------------------------------
// vector-valued case:
// -----------------------------------------------------
template<class T, class M, class BasisOnPointset>
point_basic<T>
vector_field_evaluate (
  const field_basic<T,M>&       uh,
  const BasisOnPointset&        basis_on_pointset,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof_tab,
  size_t                        q)
{
  typedef typename field_basic<T,M>::size_type size_type;
  size_type dis_ndof = uh.dis_ndof();
  size_type loc_ndof = dis_idof_tab.size();
  size_type loc_comp_ndof = std::distance (
		basis_on_pointset.begin(hat_K, q),
          	basis_on_pointset.end  (hat_K, q));
  size_type loc_comp_idof = 0;
  size_type n_comp = uh.get_geo().dimension();
  point_basic<T> value (0,0,0);
  for (typename BasisOnPointset::const_iterator
	  iter = basis_on_pointset.begin(hat_K, q),
          last = basis_on_pointset.end  (hat_K, q);
          iter != last;
          iter++, loc_comp_idof++) {
    const T& cdof = *iter;
    for (size_type i_comp = 0; i_comp < n_comp; i_comp++) {
      size_type loc_idof = loc_comp_idof + i_comp*loc_comp_ndof;
      assert_macro (loc_idof < loc_ndof, "invalid local index "<<loc_idof<<" out of range [0:"<<loc_ndof<<"[");
      size_type dis_idof = dis_idof_tab[loc_idof];
      assert_macro (dis_idof < dis_ndof, "invalid distr index");
      const T& udof = uh.dis_dof (dis_idof);
      value[i_comp] += cdof*udof;
    }
  }
  return value;
}
// -----------------------------------------------------
// tensor-valued case:
// -----------------------------------------------------
template<class T, class M, class BasisOnPointset>
tensor_basic<T>
tensor_field_evaluate (
  const field_basic<T,M>&       uh,
  const BasisOnPointset&        basis_on_pointset,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof_tab,
  size_t                        q)
{
  typedef typename field_basic<T,M>::size_type size_type;
  size_type dis_ndof = uh.dis_ndof();
  size_type loc_ndof = dis_idof_tab.size();
  size_type loc_comp_ndof = std::distance (
		basis_on_pointset.begin(hat_K, q),
          	basis_on_pointset.end  (hat_K, q));
  size_type loc_comp_idof = 0;
  size_type d = uh.get_geo().dimension();
  space_constant::coordinate_type sys_coord = uh.get_geo().coordinate_system();
  size_type n_comp = space_constant::n_component (space_constant::tensor, d, sys_coord);
  tensor_basic<T> value; // initiated to zero
  for (typename BasisOnPointset::const_iterator
	  iter = basis_on_pointset.begin(hat_K, q),
          last = basis_on_pointset.end  (hat_K, q);
          iter != last;
          iter++, loc_comp_idof++) {
    const T& cdof = *iter;
    for (size_type ij_comp = 0; ij_comp < n_comp; ij_comp++) {
      size_type loc_idof = loc_comp_idof + ij_comp*loc_comp_ndof;
      assert_macro (loc_idof < loc_ndof, "invalid local index "<<loc_idof<<" out of range [0:"<<loc_ndof<<"[");
      size_type dis_idof = dis_idof_tab[loc_idof];
      assert_macro (dis_idof < dis_ndof, "invalid distr index");
      const T& udof = uh.dis_dof (dis_idof);
      T tmp = cdof*udof;
      std::pair<size_type,size_type> ij
          = space_constant::tensor_subscript (space_constant::tensor, sys_coord, ij_comp);
      size_type i = ij.first;
      size_type j = ij.second;
      value(i,j) += tmp;
      if (i != j) value(j,i) += tmp;
    }
  }
  return value;
}
// -----------------------------------------------------
// generic interface
// -----------------------------------------------------
template<class T, class M>
inline
void
generic_field_evaluate (
  const field_basic<T,M>&       uh,
  const basis_on_pointset<T>&   bops,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof,
  size_t                        q,
  T&                            value)
{
  value = field_evaluate (uh, bops, hat_K, dis_idof, q);
}
template<class T, class M>
inline
void
generic_field_evaluate (
  const field_basic<T,M>&       uh,
  const basis_on_pointset<T>&   bops,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof,
  size_t                        q,
  point_basic<T>&               value)
{
  value = vector_field_evaluate (uh, bops, hat_K, dis_idof, q);
}
template<class T, class M>
inline
void
generic_field_evaluate (
  const field_basic<T,M>&       uh,
  const basis_on_pointset<T>&   bops,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof,
  size_t                        q,
  tensor_basic<T>&              value)
{
  value = tensor_field_evaluate (uh, bops, hat_K, dis_idof, q);
}
// -----------------------------------------------------
// homogeneous multi-component case: get the i-th value
// -----------------------------------------------------
template<class T, class M, class BasisOnPointset>
T
field_component_evaluate (
  const field_basic<T,M>&       uh,
  const BasisOnPointset&        basis_on_pointset,
  reference_element             hat_K,
  const std::vector<size_t>&    dis_idof_tab,
  size_t                        q,
  size_t                        i_comp)
{
  typedef typename field_basic<T,M>::size_type size_type;
   size_type dis_ndof = uh.dis_ndof();
   size_type loc_ndof = dis_idof_tab.size();
   size_type loc_comp_ndof = std::distance (
		basis_on_pointset.begin(hat_K, q),
          	basis_on_pointset.end  (hat_K, q));
   size_type loc_comp_idof = 0;
   size_type n_comp = uh.get_geo().dimension();
   T value = 0;
   for (typename BasisOnPointset::const_iterator
	  iter = basis_on_pointset.begin(hat_K, q),
          last = basis_on_pointset.end  (hat_K, q);
          iter != last;
          iter++, loc_comp_idof++) {
     const T& cdof = *iter;
     size_type loc_idof = loc_comp_idof + i_comp*loc_comp_ndof;
     assert_macro (loc_idof < loc_ndof, "invalid local index "<<loc_idof<<" out of range [0:"<<loc_ndof<<"[");
     size_type dis_idof = dis_idof_tab[loc_idof];
     assert_macro (dis_idof < dis_ndof, "invalid distr index");
     const T& udof = uh.dis_dof (dis_idof);
     value += cdof*udof;
   }
   return value;
}

}// namespace rheolef
#endif // _RHEOLEF_FIELD_EVALUATE_H
