#ifndef _RHEOLEF_FIELD_ON_POINTSET_H
#define _RHEOLEF_FIELD_ON_POINTSET_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================
// field_on_pointset: evaluate a field or a function on a pointset
//
#include "rheolef/geo.h"
#include "rheolef/quadrature.h"
#include "rheolef/piola.h"
#include "rheolef/field_nonlinear_expr.h"

namespace rheolef { 

// ===========================================================================
// physical representation class: a scratch array
// of all values at quadrature nodes and on all elements
// ===========================================================================
template<class T, class M>
class field_on_pointset_rep : public array<T,M> {
public:
// typedef:

  typedef array<T,M>               base;
  typedef typename base::size_type size_type;

// allocators:

  field_on_pointset_rep();
  void resize (const geo_basic<T,M>& omega, const quadrature<T>& quad, size_type n_comp=1);

// modifiers:

  template<class Expr> void set (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr);

  template<class Expr> void set_scalar_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr);

  template<class Expr> void set_vector_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr);

  template<class Expr> void set_tensor_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr);

// valued switcher:

  template<class Expr, space_constant::valued_type Tag>
  struct set_case {
   void operator() (
	field_on_pointset_rep<T,M>& obj,
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr);
  };
  template<class Expr>
  struct set_case<Expr,space_constant::scalar> {
    void operator() (
	field_on_pointset_rep<T,M>& obj,
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr) {
      obj.set_scalar_case (omega, quad, expr);
    }
  };
  template<class Expr>
  struct set_case<Expr,space_constant::vector> {
    void operator() (
	field_on_pointset_rep<T,M>& obj,
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr) {
      obj.set_vector_case (omega, quad, expr);
    }
  };
  template<class Expr>
  struct set_case<Expr,space_constant::tensor> {
    void operator() (
	field_on_pointset_rep<T,M>& obj,
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr) {
      obj.set_tensor_case (omega, quad, expr);
    }
  };

// accessors:

  space_constant::valued_type valued_tag() const { return _valued_tag; }
  std::string                 valued()     const { return space_constant::valued_name(_valued_tag); }
  size_type                   size()       const { return _valued_tag == space_constant::scalar ? 0 : _n_comp; }

  const T& at (
        const geo_size&       gs,
        const geo_element&    K,
        size_type             q,
        size_type             i_comp) const;

// data:
protected:
  boost::array<size_type,reference_element::max_variant>  _loc_nq_by_variant;
  size_type                                               _n_comp;
  space_constant::valued_type                             _valued_tag;
};
template<class T, class M>
inline
field_on_pointset_rep<T,M>::field_on_pointset_rep()
 : base(),
   _loc_nq_by_variant(),
   _n_comp(0),
   _valued_tag(space_constant::last_valued)
{
}
// ===========================================================================
// v2 interface
// ===========================================================================
// ---------------------------------------------------------------------------
// the scalar-valued case:
// ---------------------------------------------------------------------------
template<class T, class M>
template<class Expr>
void
field_on_pointset_rep<T,M>::set_scalar_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr)
{
  _valued_tag = space_constant::scalar;
  _n_comp = 1;
  resize (omega, quad, _n_comp);
  expr.initialize (omega, quad);
  std::vector<T> value;
  typename base::iterator iter = base::begin();
  for (typename geo_basic<T,M>::const_iterator
	iter_ie = omega.begin(),
	last_ie = omega.end(); iter_ie != last_ie; ++iter_ie) {
    const geo_element& K = *iter_ie;
    expr.evaluate (K, value);
    for (typename std::vector<T>::const_iterator
	iter_value = value.begin(),
	last_value = value.end(); iter_value != last_value; ++iter_value) {
      *iter++ = *iter_value;
    }
  }
}
// ---------------------------------------------------------------------------
// the vector-valued case: TODO: try to merge all valued-cases in one generic
// ---------------------------------------------------------------------------
template<class T, class M>
template<class Expr>
void
field_on_pointset_rep<T,M>::set_vector_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr)
{
  _valued_tag = space_constant::vector;
  size_type d = omega.dimension();
  _n_comp = d;
  resize (omega, quad, _n_comp);
  expr.initialize (omega, quad);
  std::vector<point_basic<T> > value;
  typename base::iterator iter = base::begin();
  for (typename geo_basic<T,M>::const_iterator
	iter_ie = omega.begin(),
	last_ie = omega.end(); iter_ie != last_ie; ++iter_ie) {
    const geo_element& K = *iter_ie;
    expr.evaluate (K, value);
    for (typename std::vector<point_basic<T> >::const_iterator
	iter_value = value.begin(),
	last_value = value.end(); iter_value != last_value; ++iter_value) {
      const point_basic<T>& tmp_value = *iter_value;
      for (size_type i_comp = 0; i_comp < _n_comp; i_comp++) {
        *iter++ = tmp_value [i_comp];
      }
    }
  }
}
// ---------------------------------------------------------------------------
// the tensor-valued case: TODO: try to merge all valued-cases in one generic
// ---------------------------------------------------------------------------
template<class T, class M>
template<class Expr>
void
field_on_pointset_rep<T,M>::set_tensor_case (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr)
{
  _valued_tag = space_constant::tensor;
  size_type d = omega.dimension();
  space_constant::coordinate_type sys_coord = omega.coordinate_system();
  _n_comp = space_constant::n_component (space_constant::tensor, d, sys_coord);
  resize (omega, quad, _n_comp);
  expr.initialize (omega, quad);
  std::vector<tensor_basic<T> > value;
  typename base::iterator iter = base::begin();
  for (typename geo_basic<T,M>::const_iterator
	iter_ie = omega.begin(),
	last_ie = omega.end(); iter_ie != last_ie; ++iter_ie) {
    const geo_element& K = *iter_ie;
    expr.evaluate (K, value);
    for (typename std::vector<tensor_basic<T> >::const_iterator
	iter_value = value.begin(),
	last_value = value.end(); iter_value != last_value; ++iter_value) {
      const tensor_basic<T>& tmp_value = *iter_value;
      for (size_type ij_comp = 0; ij_comp < _n_comp; ij_comp++) {
        std::pair<size_type,size_type> ij
	  = space_constant::tensor_subscript (space_constant::tensor, sys_coord, ij_comp);
        *iter++ = tmp_value (ij.first, ij.second);
      }
    }
  }
}
// ---------------------------------------------------------------------------
// the main call: select variant by result_type
// ---------------------------------------------------------------------------
template<class T, class M>
template<class Expr, space_constant::valued_type Tag>
inline
void
field_on_pointset_rep<T,M>::set_case<Expr,Tag>::operator() (
	field_on_pointset_rep<T,M>& obj,
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr)
{
  switch (expr.valued_tag()) {
    case space_constant::scalar: {
      obj.set_scalar_case (omega, quad, expr);
      break;
    }
    case space_constant::tensor: {
      obj.set_tensor_case (omega, quad, expr); 
      break;
    }
    default:
      fatal_macro ("unexpected "<<space_constant::valued_name(expr.valued_tag())
	<< "-valued field-like argument");
  }
}
template<class T, class M>
template<class Expr>
inline
void
field_on_pointset_rep<T,M>::set (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	const field_nonlinear_expr<Expr>&  expr)
{
  set_case <Expr, Expr::valued_hint> set_with_tag;
  set_with_tag (*this, omega, quad, expr);
}
// ===========================================================================
// smart pointer interface
// ===========================================================================
template <class T, class M>
class field_on_pointset : public smart_pointer<field_on_pointset_rep<T,M> > {
public:
// typedefs:
  typedef field_on_pointset_rep<T,M> rep;
  typedef smart_pointer<rep>         base;
  typedef typename rep::size_type    size_type;
// allocators:
  field_on_pointset() : base (new_macro(rep)) {}
  void resize (const geo_basic<T,M>& omega, const quadrature<T>& quad, size_type n_comp=1) {
    base::data().resize (omega, quad, n_comp); }
// accessor:
  space_constant::valued_type valued_tag() const { return base::data().valued_tag(); }
  std::string                 valued()     const { return base::data().valued(); }
  size_type                   size()       const { return base::data().size(); }

  const T& operator[] (size_type iq) const { return base::data().operator[] (iq); }
  const T& at (
        const geo_size&       gs,
        const geo_element&    K,
        size_type             q,
        size_type             i_comp) const
  	{ return base::data().at (gs, K, q, i_comp); }

// modifiers:
  template<class Function> void set (
  	const geo_basic<T,M>&       omega,
	const quadrature<T>&        quad,
	Function                    f)
  	{ base::data().set (omega, quad, f); }
};

}// namespace rheolef
# endif // _RHEOLEF_FIELD_ON_POINTSET_H
