# ifndef _RHEOLEF_FORM_WEIGHTED_H
# define _RHEOLEF_FORM_WEIGHTED_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================
// form with a weight as a class-function
#include "rheolef/form.h"
namespace rheolef { 

// ---------------------------------------------------------------------------
// form cstor
// ---------------------------------------------------------------------------
template<class T, class M>
template<class WeightFunction>
void
form_basic<T,M>::form_init (
    const std::string&      name,
    bool                    has_weight,
    WeightFunction          weight,
    const quadrature_option_type& qopt)
{
    if (name == "" || name == "nul" || name == "null") {
      // empty name => nul form, but with declared csr matrix sizes
      _uu.resize (_Y.iu_ownership(), _X.iu_ownership());
      _ub.resize (_Y.iu_ownership(), _X.ib_ownership());
      _bu.resize (_Y.ib_ownership(), _X.iu_ownership());
      _bb.resize (_Y.ib_ownership(), _X.ib_ownership());
      return;
    }
    typedef typename form_basic<T,M>::float_type float_type;
    bool X_is_on_domain = (_X.get_geo().data().variant() == geo_abstract_base_rep<float_type>::geo_domain);
    bool Y_is_on_domain = (_Y.get_geo().data().variant() == geo_abstract_base_rep<float_type>::geo_domain);
    geo_basic<T,M> omega;
    if (X_is_on_domain == Y_is_on_domain) {
      omega = _X.get_geo();
    } else {
      omega = _X.get_geo().get_background_geo();
    }
    form_element<T,M> form_e (name, _X, _Y, omega, qopt);

    if (has_weight) form_e.set_weight (weight);

    if (X_is_on_domain == Y_is_on_domain) {
      // example:
      //    form m (V,V,"mass");    e.g. \int_\Omega u v dx
      //    form m (W,W,"mass");    e.g. \int_\Gamma u v ds
      // with:
      //    geo omega ("square");
      //    geo gamma = omega["boundary"];
      //    V = space(omega,"P1");
      //    W = space(gamma,"P1");
      assembly (form_e, _X.get_geo(), _Y.get_geo());
    } else if (X_is_on_domain) {
      // example:
      //    form m (W,V,"mass");   e.g. \int_\Gamma u trace(v) ds
      check_macro (_X.get_geo().get_background_geo() == _Y.get_geo(), 
	"form between incompatible geo " << _X.get_geo().name() << " and " << _Y.get_geo().name());
      const geo_basic<T,M>& Y_gamma = _X.get_geo().get_background_domain();
      bool X_geo_is_background = false;
      assembly (form_e, _X.get_geo(), Y_gamma, X_geo_is_background);
    } else { // Y_is_on_domain
      // example:
      //    form m (V,W,"mass");  e.g. \int_\Gamma trace(u) v ds	
      check_macro (_Y.get_geo().get_background_geo() == _X.get_geo(),
	"form between incompatible geo " << _X.get_geo().name() << " and " << _Y.get_geo().name());
      const geo_basic<T,M>& X_gamma = _Y.get_geo().get_background_domain();
      bool X_geo_is_background = true;
      assembly (form_e, X_gamma, _Y.get_geo(), X_geo_is_background);
    }
}
template<class T, class M>
template<class Function>
inline
form_basic<T,M>::form_basic (
  const space_type& X, 
  const space_type& Y,
  const std::string& name,
  Function weight,
  const quadrature_option_type& qopt)
  : _X(X),
    _Y(Y),
    _uu(),
    _ub(),
    _bu(),
    _bb()
{
    form_init (name, true, weight, qopt);
}

}// namespace rheolef
# endif /* _RHEOLEF_FORM_WEIGHTED_H */
